/*
	GapList: class for managing free space in swap files.
	
	Handles anything to do with creating, consolidating, truncating, deleting, and
	finding runs of free space.
*/
#include "GapList.h"
#include <OS.h>

class Gap
{
public:
	Gap(uint64 s, uint64 e) { start=s; end=e; next=prev=NULL; }
	uint64 GetSize(void) { return (end>start)?(end-start):0; }

	uint64 start;
	uint64 end;

	Gap *prev;
	Gap *next;
};

GapList::GapList(const uint64 &startingsize)
{
	head=tail=NULL;
	itemcount=0;
	poolsize=startingsize;
}

GapList::~GapList(void)
{
	Gap *current=head, *temp;

	while(current)
	{
		temp=current->next;
		delete current;
		current=temp;
	}
}

// Add a run to the list of free space
void GapList::Add(const uint64 &start, const uint64 &end)
{

	if(start>=end || start>poolsize)
		return;
	
	uint64 gstart=start,gend=end;

	if(gend>poolsize)
		gend=poolsize;
	
	Gap *index=head, *prev=NULL, *next=NULL;
	
	// Search for the item which should go before the one we are adding
	while(index)
	{
		if(index->next && index->next->start>gstart)
		{
			prev=index;
			break;
		}
		index=index->next;
	}
	
	// Search for the item which should go after the one we are adding
	index=prev;
	while(index)
	{
		if(index->next && index->next->end>gend)
		{
			next=index->next;
			break;
		}
		index=index->next;
	}
	
	if(!prev)
	{
		// If prev==NULL, that means one of two things - either there aren't any
		// items in the list at all or it needs to be placed at the end of the list
		if(itemcount==0)
		{
			tail=head=new Gap(gstart,gend);
		}
		else
		{
			Gap *newtail=new Gap(gstart,gend);
			tail->next=newtail;
			newtail->prev=tail;
			tail=newtail;
		}
		return;
	}


	// we got this far so the gap values are legitimate and we only have to be concerned
	// with overlaps
	
	// There are five cases to account for:
	// 1) new gap inside an existing gap
	// 2) new gap envelops one or more existing gaps
	// 3) new gap overlaps one or two existing gaps
	// 4) cases 2 and 3 combined
	// 5) None of the above - just add the stupid thing :P
	Gap *newgap;
	
	if(gstart>prev->end && gend<next->start)
	{
		// case 5
		newgap=new Gap(gstart,gend);
		prev->next=newgap;
		next->prev=newgap;
		newgap->prev=prev;
		newgap->next=next;
		return;
	}
	
	if(gend<prev->end)
	{
		// case 1: if completely contained by another gap, we don't have to do anything
		if(gstart>prev->start)
			return;
		
		// case 3
		if(gend<next->start)
		{
			prev->end=gend;
			return;
		}
		
		// case 4 - the really annoying case
		Gap *lastinrun=index=next;

		while(index)
		{
			if(index->end>gend)
			{
				lastinrun=index;
				break;
			}
			index=index->next;
		}
		
		// delete any gaps completely enveloped
		index=prev->next;
		while(index!=lastinrun)
		{
			Gap *tempgap=index;
			tempgap->prev->next=tempgap->next;
			tempgap->next->prev=tempgap->prev;
			index=index->next;
			delete tempgap;
		}
		
		// Theoretically, we now only deal with two overlapping gaps
		prev->end=lastinrun->end;
		delete lastinrun;
		return;
	}

	// Having reached this point in the function, it means that we do not overlap
	// the previous item, but we *do* overlap one after the gap

	Gap *lastinrun=index=next;

	while(index)
	{
		if(index->end>gend)
		{
			lastinrun=index;
			break;
		}
		index=index->next;
	}
	
	// delete any gaps completely enveloped
	index=next;
	while(index!=lastinrun)
	{
		Gap *tempgap=index;
		tempgap->prev->next=tempgap->next;
		tempgap->next->prev=tempgap->prev;
		index=index->next;
		delete tempgap;
	}
	
	// Theoretically, we only have to resize the gap after where we want to insert the gap
	lastinrun->start=gstart;
}

// Appends a block of listed size to existing free space
void GapList::Expand(const uint64 &size)
{
	
	if(tail && tail->end==poolsize)
		tail->end+=size;

	poolsize+=size;
}

// Truncates the list of free space to fit within the given size
void GapList::Truncate(const uint64 &newsize)
{
	// Remove any gaps past the new EOF
	while(tail->start>newsize)
	{
		tail->prev->next=NULL;

		Gap *oldtail=tail;
		
		RemoveItem(oldtail);
		delete oldtail;
	}
		
	// If the new EOF is in the middle of the gap, then we simply
	// need to resize it and we are done
	if(tail->end>newsize)
		tail->end=newsize;

	poolsize=newsize;
}

// Obtains the location of a block of the given size and removes it
// from the list of free space
status_t GapList::Get(const uint32 &size, uint64 &position)
{
	if(itemcount==0 || size==0)
		return B_ERROR;

	if(size>poolsize)
		return B_BAD_VALUE;
	
	Gap *index=head;

	while(index)
	{
		if(index->GetSize()>size)
		{
			position=index->start;
			index->start+=size;
			return B_OK;
		}

		if(index->GetSize()==size)
		{
			position=index->start;
			Gap *temp=index;
			index=index->next;			
			RemoveItem(temp);
			delete temp;
			return B_OK;
		}

		if(index->GetSize()<=0)
		{
			Gap *temp=index;
			index=index->next;			
			RemoveItem(temp);
			delete temp;
			continue;
		}
		index=index->next;
	}
		
	return B_NO_MEMORY;
}

// Show the position of the first available free space. Note that the space is
// not removed from the list
status_t GapList::SeeFirst(uint64 &position) const
{
	if(!head)
		return B_ERROR;

	position=head->start;
	return B_OK;
}

// Show the position of the last available free space. Note that the space is
// not removed from the list
status_t GapList::SeeLast(uint64 &position) const
{
	if(!tail)
		return B_ERROR;

	position=tail->start;
	return B_OK;
}

// Show the position of a particular gap. Note that the space is
// not removed from the list
status_t GapList::SeeItem(const uint32 &index, uint64 &position) const
{
	if(index>itemcount-1)
		return B_ERROR;
	
	Gap *temp=head;

	for(uint32 i=0; i<index; i++)
	{
		temp=temp->next;
		if(!temp)
			return B_ERROR;
	}
	position=temp->start;
	return B_OK;
}

// Removes (but not deletes) an item from the list
void GapList::RemoveItem(Gap *gap)
{
	if(itemcount==0 || !gap)
		return;
	
	if(itemcount==1)
	{
		head=NULL;
		tail=NULL;
		itemcount=0;
		return;
	}
	
	// remove from start of list
	if(head==gap)
	{
		head=gap->next;
		head->prev=NULL;
		
		itemcount--;
		return;
	}

	// remove from end of list
	if(tail==gap)
	{
		tail=tail->prev;
		tail->next=NULL;
		itemcount--;
		return;
	}
	
	// remove from middle of list
	if(!gap->prev && !gap->next)
		debugger("Gap in the middle of list with NULL pointers");
	
	gap->prev->next=gap->next;
	gap->prev->next->prev=gap->prev;
	itemcount--;
}
