/*
	TileCache.cpp:	Object to cache tile accesses for greater speed
*/
#include "TileCache.h"
#include "TileVMM.h"

//#define DEBUG_CACHE
//#define LOG_CACHE

#ifdef DEBUG_CACHE
#include <stdio.h>
#define STRACE(a) printf a

#ifdef LOG_CACHE
#define SLOG(a) printf a
#else
#define SLOG(a) printf  /* nothing */
#endif

#else
#define STRACE(a) /* nothing */

#ifdef LOG_CACHE
#define SLOG(a) printf a
#else
#define SLOG(a) /* nothing */
#endif

#endif

uint64 default_cache_size=DEFAULT_CACHE_SIZE;

uint64 get_default_cache_size(void)
{
	return default_cache_size;
}

void set_default_cache_size(uint64 size)
{
	default_cache_size=size;
}

TileCache::TileCache(void)
{
	SLOG(("TileCache::TileCache()\n"));

	currentsize=0;
	maxsize=get_default_cache_size();
}

TileCache::~TileCache(void)
{
	SLOG(("TileCache::~TileCache()\n"));
	Tile *t=tilelist.GetOldest();

	// Because there's only one cache to a VMM, we can afford to do this
	if(t)
		t->SetDirty(false);
	while(t)
	{
		FlushTile(t);
		t->SetDirty(false);
		
		t=tilelist.GetOldest();
	}
}

void TileCache::SetMaxSize(uint64 size)
{
	SLOG(("TileCache::SetMaxSize(%lld)\n",size));
	Tile *t=tilelist.GetOldest();
	
	if(size<currentsize)
	{
		do
		{
			t=tilelist.GetOldest();
			FlushTile(t);
		} while(t);
	}

	maxsize=size;
}

bool TileCache::HasTile(Tile *tile)
{
	SLOG(("TileCache::HasTile(%p)\n",tile));
	if(!tile || !tile->GetData())
		return false;
	
	Tile *t;
	
	Lock();
	t=tilelist.Get(tile->GetID());
	Unlock();

	return (t!=NULL);
}

bool TileCache::TileWillFit(void)
{
	bool value=true;
	
	Lock();
	if( (currentsize + MAX_TILE_SIZE) > maxsize )
		value=false;
	Unlock();
	return value;
}

status_t TileCache::AddTile(Tile *tile)
{
	SLOG(("TileCache::AddTile(%p), size before: %lld\n",tile,currentsize));
	if(!tile)
		return B_ERROR;

	if(!tile->GetData())
	{
		STRACE(("Invalid tiles may not be cached\n"));
		return B_BAD_VALUE;
	}
	
	Lock();
	
	// Look to see if the tile is already cached. If it is, remove it from the list
	// so that we can add it at the beginning of the list (most recently used)
	if(tilelist.Get(tile->GetID()))
	{
		tilelist.Remove(tile->GetID());
		currentsize-=MAX_TILE_SIZE;
	}
	
	// If the tile won't fit into the cache, it's up to the owning TileVMM to do
	// something about it.
	if( (currentsize + MAX_TILE_SIZE) > maxsize )
	{
		Unlock();
		return B_NO_MEMORY;
	}
			
	// Add the tile to the list and increment the cache's size
	tilelist.Add(tile);
	currentsize+=MAX_TILE_SIZE;
	
	Unlock();

	return B_OK;
}

status_t TileCache::FlushTile(Tile *tile)
{
	SLOG(("TileCache::FlushTile(%p)\n",tile));

	if(!tile)
		return B_ERROR;
	
	// The tile passed to us can't possibly be cached because it's invalid.
	// We don't cache invalid tiles.
	if(!tile->IsValid())
		return B_NO_MEMORY;
	
	// We're not allowed to flush tiles which haven't been synced with the disk
	if(tile->IsDirty())
		return B_NOT_ALLOWED;
	
	Lock();

	currentsize-=MAX_TILE_SIZE;
	tilelist.Remove(tile->GetID());

	Unlock();
	
	return B_OK;
}

Tile *TileCache::GetOldestTile(void)
{
	SLOG(("TileCache::FlushOldest\n"));

	Tile *tile;

	Lock();
	tile=tilelist.GetOldest();
	Unlock();
	
	return tile;
}

Tile *TileCache::GetTile(int32 id)
{
	SLOG(("TileCache::GetTile(%ld)\n",id));
	Lock();
	
	Tile *tile=NULL;
	
	tile=tilelist.Get(id);
	if(tile)
	{
		tilelist.Remove(tile->GetID());
		tilelist.Add(tile);
	}

	Unlock();
	return tile;
}
