/*
	TileManager.cpp:	Class to track tiles in a bitmap
*/
#include "TileManager.h"

#include "TileCache.h"
#include "TileVMM.h"

#define DEBUG_MANAGER
#define LOG_MANAGER

#ifdef DEBUG_MANAGER
#include <stdio.h>
#define STRACE(a) printf a

#ifdef LOG_MANAGER
#define SLOG(a) printf a
#else
#define SLOG(a) /* nothing */
#endif

#else
#define STRACE(a) /* nothing */

#ifdef LOG_MANAGER
#define SLOG(a) printf a
#else
#define SLOG(a) /* nothing */
#endif

#endif

extern uint32 global_tile_count;
extern uint32 active_tile_count;
extern uint32 shared_tile_count;

extern TileVMM *global_tile_vmm;

TileManager::TileManager(int16 w, int16 h, int16 bytespp)
{
	SLOG(("TileManager:TileManager(%d,%d,%d\n",w,h,bytespp));
	width=w;
	height=h;
	bpp=bytespp;
	rowcount=(h+MAX_TILE_HEIGHT-1)/ MAX_TILE_HEIGHT;
	colcount=(w+MAX_TILE_WIDTH-1)/ MAX_TILE_WIDTH;
	tiles=NULL;

	vmm=global_tile_vmm;

	AllocateTiles(rowcount * colcount);
}

TileManager::~TileManager(void)
{
	SLOG(("TileManager~TileManager()\n"));
	if(tiles)
	{
		int16 tilecount=rowcount*colcount;
		for(int16 i=0; i<tilecount; i++)
		{
			tiles[i]->Lock();
			DetachTile(tiles[i],i);
		}
		
		delete [] tiles;
	}
}

Tile *TileManager::GetTile(int16 x, int16 y)
{
	int16 id=GetTileID(x,y);
	if(id<0)
		return NULL;
	
	return GetTile(id);
}

Tile *TileManager::GetTile(const int32 &tileid)
{
	SLOG(("TileManager::GetTile(%ld)\n",tileid));
	int tilecount;
	Tile *requested_tile;
	
	tilecount=rowcount * colcount;
	
	if ((tileid < 0) || (tileid >= tilecount))
		return NULL;
	
	requested_tile=vmm->GetTile(tileid);
	
	if(!requested_tile)
		return NULL;
	
//	if(!requested_tile->IsValid())
//		vmm->Validate(requested_tile);

	if(requested_tile->sharecount>1)
	{
		Tile *newtile=vmm->DuplicateTile(requested_tile);
		DetachTile(requested_tile,tileid);
		AttachTile(newtile,tileid);
		requested_tile=newtile;
	}
	
	requested_tile->SetDirty(true);
	return requested_tile;
}

int32 TileManager::GetTileID(int16 x, int16 y)
{
	int16 row,col,id;
	
	if ((x<0) || (x>=width) || (y<0) || (y>=height))
		return -1;
	
	row= y/MAX_TILE_HEIGHT;
	col= x/MAX_TILE_WIDTH;
	id=(row * colcount) + col;
	
	return id;
}

void TileManager::MapTile(int16 x, int16 y, Tile *source)
{
	int16 row,col,id;
	
	if((x<0) || (x>=width) || (y<0) || (y>=height))
	{
		STRACE(("Coordinate out of bounds in TileManager::MapTile()\n"));
		return;
	}
	
	row= y/MAX_TILE_HEIGHT;
	col= x/MAX_TILE_WIDTH;
	id=row * colcount + col;
	
	MapTile(id,source);
}

void TileManager::MapTile(int32 tileid, Tile *source)
{
	SLOG(("TileManager::MapTile(%ld,%p)\n",tileid,source));
	int16 tilecount=rowcount*colcount;
	Tile **tileptr;
	
	if( (tileid<0) || (tileid>tilecount))
	{
		STRACE(("tile ID out of bounds in TileManager::MapTile\n"));
		return;
	}
	
/*	if(!tiles)
	{
		STRACE(("TileManager::MapTile - NULL tile pointer. Initializing...\n"));
		AllocateTiles(tilecount);
	}
*/	
	tileptr=&tiles[tilecount];
	
	if(!source->IsValid())
		debugger("Invalid tiles may not be mapped");
	
	(*tileptr)->Lock();
	
	if( (*tileptr)->width!=source->width ||
		(*tileptr)->height!=source->height ||
		(*tileptr)->bpp!=source->bpp )
	{
		STRACE(("Unequal mapping in TileManager::MapTile\n"));
	}
	DetachTile(*tileptr,tileid);
	
	source->Lock();	
	AttachTile(source,tileid);
	*tileptr=source;
	source->Unlock();
}

void TileManager::MapOverTile(Tile *source, Tile *dest)
{
	TileLink *tlink;
	for(tlink=dest->links; dest!=NULL; tlink=tlink->next)
		if(tlink->manager==this) break;
	
	if(!tlink)
	{
		STRACE(("Destination tile not attached to manager in TileManager::MapOverTile\n"));
		return;
	}
	MapTile(tlink->id, source);
}

void TileManager::ValidateTile(Tile *tile)
{
	SLOG(("TileManager::ValidateTile(%p)\n",tile));
	if(!tile)
	{
		STRACE(("NULL tile passed to TileManager::Validate\n"));
		return;
	}
//	tile->valid=true;
	vmm->Validate(tile);
}

void TileManager::Invalidate(Tile **tileptr,int16 tileid)
{
	SLOG(("TileManager::Invalidate(%p,%d)\n",tileptr,tileid));
	if(!tileptr)
	{
		STRACE(("NULL Tile ** in TileManager::Invalidate()\n"));
		return;
	}
	
	Tile *tile=*tileptr;

	tile->Lock();
	
	if(!tile->IsValid())
	{
		tile->Unlock();
		return;
	}
	
	if(tile->sharecount>1)
	{
		Tile *newtile=vmm->DuplicateTile(tile);
		DetachTile(tile,tileid);
		newtile->Lock();
		AttachTile(newtile,tileid);
		tile=*tileptr=newtile;
	}
	
	vmm->Invalidate(tile);
	
	tile->Unlock();
}

void TileManager::Invalidate(Tile **tileptr,int16 x, int16 y)
{
	int16 id=GetTileID(x,y);
	if(id<0)
		return;

	Invalidate(tileptr, id);
}

void TileManager::InvalidateTiles(Tile *toplevel)
{
	double x,y;
	int16 row,col,id;
	
	col=toplevel->links->id % colcount;
	row=toplevel->links->id / colcount;
	
	x = (col * MAX_TILE_WIDTH + toplevel->GetWidth() / 2.0) / (double) width;
	y = (row * MAX_TILE_HEIGHT + toplevel->GetHeight() / 2.0) / (double) height;
	
	if(tiles)
	{
		col=int16(x* width/MAX_TILE_WIDTH);
		row=int16(y* height/MAX_TILE_HEIGHT);
		id=row * colcount + col;
		Invalidate(&tiles[id],id);
	}
}

void TileManager::InvalidateTile(Tile *tile)
{
	if(!tile->IsValid())
		return;
	
	vmm->Invalidate(tile);
}

void TileManager::GetTileCoordinates(Tile *tile, int16 *x, int16 *y)
{
	TileLink *tlink;
	for (tlink=tile->links; tlink!=NULL; tlink=tlink->next) 
		if (tlink->manager==this) break;
	
	if (!tlink) 
	{
		STRACE(("tile_manager_get_tile_coordinates: tile not attached to manager"));
		return;
	}
	
	*x = MAX_TILE_WIDTH * (tlink->id % colcount);
	*y = MAX_TILE_HEIGHT * (tlink->id / colcount);
}

void TileManager::AttachTile(Tile *tile, int16 tileid)
{
//	SLOG(("TileManager::AttachTile(%p,%d)\n",tile,tileid));
	if( (tile->sharecount>0) && !tile->IsValid())
		vmm->Validate(tile);
	tile->sharecount++;
	shared_tile_count++;
	
	TileLink *link=new TileLink;
	link->manager=this;
	link->id=tileid;
	link->next=tile->links;
	tile->links=link;
}

void TileManager::DetachTile(Tile *tile, int16 tileid)
{
//	SLOG(("TileManager::DetachTile(%p,%d)\n",tile,tileid));
	TileLink **link, *tmp;
	
	for(link=&tile->links; *link!=NULL; link=&(*link)->next )
	{
		if( ((*link)->manager==this) &&
				((*link)->id==tileid) )
			break;
	}
	if(*link==NULL)
	{
		STRACE(("Attempted to detach a nonattached tile in TileManager::DetachTile\n"));
		return;
	}
	tmp=*link;
	*link=tmp->next;
	delete tmp;
	
	if(tile->sharecount==0 && tile->refcount==0)
		delete tile;

	tile->Unlock();
}
#include <stdio.h>
void TileManager::AllocateTiles(int32 tilecount)
{
	SLOG(("TileManager::AllocateTiles(%ld)\n",tilecount));
	int16 right_tile_width, bottom_tile_height;
	int16 i,j;
	uint64 k;
	
	// Create the pointer array for the list of tiles
	tiles=new Tile*[tilecount];
	
	right_tile_width=width - ((colcount - 1) * MAX_TILE_WIDTH);
	bottom_tile_height=height - ((rowcount - 1) * MAX_TILE_HEIGHT);
	
	for (i=0, k=0; i<rowcount; i++)
	{
		for (j=0; j<colcount; j++, k++)
		{
//			tiles[k]=new Tile(bpp);
//			tiles[k]->id=k;
			tiles[k]=vmm->CreateTile(bpp);
			AttachTile(tiles[k],k);
			
			if(j==(colcount-1))
				tiles[k]->width=right_tile_width;
			
			if(i==(rowcount-1))
				tiles[k]->height=bottom_tile_height;
		}
	}
}
