/*
	TileVMM.cpp:	Memory manager object for BPM
*/
#include "TileVMM.h"
#include <List.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <File.h>
#include "DebugTools.h"
#include "GapList.h"

//#define DEBUG_VMM
//#define LOG_VMM

#ifdef DEBUG_VMM
#define STRACE(a) printf a

#ifdef LOG_VMM
#define SLOG(a) printf a
#else
#define SLOG(a) /* nothing */
#endif

#else
#define STRACE(a) /* nothing */

#ifdef LOG_VMM
#define SLOG(a) printf a
#else
#define SLOG(a) /* nothing */
#endif

#endif

#define MAX_OPEN_FILES 16

#define SWAPFILE_BLOCK_SIZE (MAX_TILE_WIDTH * MAX_TILE_HEIGHT * 64)

uint32 global_swap_counter=0;
TileVMM *global_tile_vmm=NULL;

int32 tile_token_index=0;

class SwapFile : public BFile
{
public:
	SwapFile(const char *filename);
	~SwapFile();
	void SetID(int16 sfid) { id=sfid; }
	int16 GetID(void) const { return id; }
	const char *GetName(void) const { return name.String(); }
	off_t GetPosition(void) const { return position; }
	void SetPosition(off_t pos) { position=pos; }
//	int GetFD(void) const { return fd; }
//	void SetFD(int newfd) { fd=newfd; }
	GapList &GetGaplist(void) { return gaps; }
	uint64 GetFileSize(void) const { return size; }
	void SetFileSize(uint64 newsize);
protected:
	BString name;
	int16 id;
//	int fd;
	GapList gaps;
	uint64 size;
	uint64 position;
};

SwapFile::SwapFile(const char *filename)
 : BFile(filename, B_CREATE_FILE | B_READ_WRITE | B_ERASE_FILE),
 gaps(0)
{
	SLOG(("SwapFile:SwapFile(%s)\n",filename));
	if(filename)
		name.SetTo(filename);
	id=global_swap_counter++;

//	fd=open(filename,O_CREAT | O_BINARY | O_RDWR);
	
	size=0;
	position=0;
}

SwapFile::~SwapFile(void)
{
	SLOG(("SwapFile:~SwapFile()\n"));
//	if(fd!=-1)

	if(InitCheck()==B_OK)
	{
//		close(fd);
		Unset();
		unlink(name.String());
	}
}

void SwapFile::SetFileSize(uint64 newsize)
{
	// TODO: Do this right - if file is expanding, add or expand a gap. If file is
	// shrinking set file size and cull invalid gaps
	
/*	if(size<newsize)
	{
		// we are making the file bigger, so we need to do one of two things:
		// get the last gap in the file and extend it to the new end of file OR
		// create a new gap at the end of the file if there isn't one
		
		Gap *gap;
		for(int32 i=0; gaps.CountItems(); i++)
		{
			gap=(Gap*)gaps.ItemAt(i);
			if(!gap)
				continue;
			if(gap->end==size)
			{
				// we found a gap and it ended at the file's end. Set its
				// new end to the new EOF
				gap->end=newsize;
				size=newsize;
				BFile::SetSize(newsize);
				return;
			}
		}
		
		// Got this far, so add a gap
		gaps.AddItem(new Gap(size,newsize));
	}
*/
	if(size>newsize)
		gaps.Truncate(newsize);
	else
		gaps.Expand(newsize);
	
	size=newsize;
	BFile::SetSize(newsize);
}

TileVMM::TileVMM(const char *path)
{
	SLOG(("TileVMM::TileVMM\n"));
	vmcache=new TileCache;

	char tempname[B_PATH_NAME_LENGTH];

	sprintf(tempname,"%s/BPMTempFileXXXXXX",path);
	mktemp(tempname);
	
	// Create a swap file for the VMM to store all tiles in.
	CreateSwapFile(tempname);
}

TileVMM::~TileVMM(void)
{
	SLOG(("TileVMM::~TileVMM\n"));
	delete vmcache;
	delete swapfile;
}

Tile *TileVMM::CreateTile(int16 bytesperpixel)
{
	// Called to instantiate a tile object
	
	Lock();
	
	// Currently, the only supported color spaces are RGB(A), grayscale, and indexed
	if(bytesperpixel!=4 && bytesperpixel!=1)
		return NULL;
	
	Tile *tile=new Tile(bytesperpixel);
	tile->id=tile_token_index++;
	tilelist.Add(tile);
	Validate(tile);
	
	while(!vmcache->TileWillFit())
	{
		// If there isn't room in the cache for another tile, get the least-recently-accessed
		// tile and swap it out to disk
		Tile *oldest=vmcache->GetOldestTile();
		SwapTileOut(oldest);
	}
	
	vmcache->AddTile(tile);
	Unlock();
	
	return tile;
}

void TileVMM::DestroyTile(int32 tileid)
{
	// Called to delete a tile object
	Lock();
	
	Tile *tile=tilelist.Get(tileid);

	if(tile)
	{
		if(vmcache->GetTile(tileid))
			vmcache->FlushTile(tile);
			
		if(tile->IsSwapped())
			RemoveTileFromSwap(tile);
		
		tilelist.Remove(tile->GetID());
		
		delete tile;
	}
	
	Unlock();
}

Tile *TileVMM::DuplicateTile(Tile *tile)
{
	if(!tile)
		return NULL;
	
	Lock();
	
	Tile *newtile=new Tile(tile->bpp);
	*newtile=*tile;
	newtile->id=tile_token_index++;
	tilelist.Add(newtile);
	vmcache->AddTile(tile);
	Unlock();
	return newtile;
}

Tile *TileVMM::GetTile(const int32 &tileid)
{
	Lock();
	
	// Check the cache for its existence;
	Tile *tile=vmcache->GetTile(tileid);
	
	if(tile)
	{
		SLOG(("TileVMM::GetTile(): Get tile from cache\n"));
		Unlock();
		return tile;
	}
	
	// Apparently, we have a cache miss, so swap the thing in.
	SLOG(("TileVMM::GetTile(): Get cache miss\n"));
	tile=tilelist.Get(tileid);
	SwapTileIn(tile);
	
	Unlock();
	return tile;
}

void TileVMM::Validate(Tile *tile)
{
	if(!tile || tile->GetData())
		return;
	
	Lock();
	tile->Validate();
	Unlock();
}

void TileVMM::Invalidate(Tile *tile)
{
	if(!tile || !tile->GetData())
		return;
	
	Lock();

	if(vmcache->HasTile(tile))
		vmcache->FlushTile(tile);
	tile->Invalidate();

	Unlock();
}

int16 TileVMM::CreateSwapFile(const char *filename)
{
	SLOG(("TileVMM::CreateSwapFile(%s)\n",filename));
	if(!filename)
		return -1;

	Lock();

	// For now, this isn't necessary because we're only using 1 swap file.
		
//	SwapFile *file;
//	file=new SwapFile(filename);
	
	// I think this is the right way to add it to the table...
//	swaplist[file->GetID()]=file;
	
	swapfile=new SwapFile(filename);

	Unlock();
	return swapfile->GetID();
}

void TileVMM::DeleteSwapFile(int16 swapid)
{
	SLOG(("TileVMM::DeleteSwapFile(%d)\n",swapid));
	Lock();
/*	SwapFile *file=swaplist[swapid];

	if(file)
	{
		swaplist.erase(swapid);
		delete file;
	}
*/
	if(swapfile)
	{
		delete swapfile;
		swapfile=NULL;
	}
		
	Unlock();
}

void TileVMM::RemoveTileFromSwap(Tile *tile)
{
	SLOG(("TileVMM::DeleteTile(%p)\n",tile));
	if(!tile || !tile->IsSwapped())
		return;
	
//	SwapFile *file=swaplist[tile->GetSwapID()];
//	if(!file)
//		return;

	SwapFile *file=swapfile;

	if(!swapfile)
	{
		STRACE(("ERROR: Attempted to swap out a tile without a swap file in TileVMM!\n"));
		return;
	}

	uint32 start=tile->GetSwapOffset();
	uint32 end=start+(MAX_TILE_WIDTH * MAX_TILE_HEIGHT * tile->BytesPerPixel());

	GapList &gaplist=file->GetGaplist();
	gaplist.Add(start,end);
	
	tile->SetSwapOffset(-1);
	
/*	Gap *gap=NULL, *tempgap=NULL;
	
	for(int32 i=0; i<gaplist.CountItems(); i++)
	{
		gap=(Gap*)gaplist.ItemAt(i);

		if(end==gap->start)
		{
			gap->start=start;

			if(i>0)
			{
				tempgap=(Gap*)gaplist.ItemAt(i-1);
				if(tempgap && gap->start==tempgap->end)
				{
					// There is a gap right before the current tile, so expand it
					// to encompass the whole area and delete the curent gap 
					tempgap->end=gap->end;

					gaplist.RemoveItem(gap);
					delete gap;
				}
			}
			
			break;
		}
		else
			if(start==gap->end)
			{
				gap->end=end;
				
				if(i< (gaplist.CountItems()-2) )
				{
					tempgap=(Gap*)gaplist.ItemAt(i+1);
					if(tempgap)
					{
						if(gap->end==tempgap->start)
						{
							// There is a gap immediately following the current tile, so expand it
							// to encompass the whole area and delete the current gap 
							tempgap->start=gap->start;
							
							gaplist.RemoveItem(gap);
							delete gap;
						}
					}
				}
				break;
			}
			else
				if(end < gap->start)
				{
					gap=new Gap(start,end);
					gaplist.AddItem(gap);
					break;
				}
				else
					if(i==gaplist.CountItems()-1)
					{
						gap=new Gap(start, end);
						gaplist.AddItem(gap);
					}
	}
	
	if(gaplist.IsEmpty())
	{
		gap=new Gap(start,end);
		gaplist.AddItem(gap);
	}
*/


/*	if(gap->end==file->GetFileSize())
	{
		ResizeFile(file,gap->start);
		gaplist.RemoveItem(gap);
		delete gap;
	}
*/
	uint64 pos;
	if(gaplist.SeeLast(pos)==B_OK)
	{
		if(pos==file->GetFileSize())
		{
			ResizeFile(file,pos);
			gaplist.Truncate(pos);
		}
	}
}

void TileVMM::SwapTileOut(Tile *tile)
{
	// Send a tile out to the disk. Invalid tiles may not be paged out because there would
	// be no data to write there. This function is also responsible to ensure that new
	// swap files are created when they are needed.
	
	SLOG(("TileVMM::SwapTileOut(%p)\n",tile));
	if(!tile || !tile->IsValid())
		return;
	
	Lock();
	
	// First, find out which file to write to. I can't see the advantage to having
	// more than one swapfile open at a time, so we'll just use the one created in the
	// constructor.
		
	SwapFile *file=swapfile;
	
	int writeerror=0;
	off_t filepos=-1;
	
	if(tile->IsSwapped())
		filepos=tile->GetSwapOffset();
	else
		filepos=ObtainOffset(file,tile->BitsLength());
			
	writeerror=file->WriteAt(filepos, tile->GetData(), tile->BitsLength());
	if(writeerror<=0)
		printf("Couldn't write file in SwapTileIn\n");
	
	tile->SetDirty(false);
	tile->SetSwapOffset(filepos);

	vmcache->FlushTile(tile);
	
	tile->Invalidate();
	tile->swapped=true;
	
	Unlock();
}

void TileVMM::SwapTileIn(Tile *tile)
{
	SLOG(("TileVMM::SwapTileIn(%p)\n",tile));
	if(!tile || !tile->IsSwapped() || tile->IsValid())
		return;
	
	Lock();
	
//	SwapFile *file=swaplist[tile->swapfile_id];
//	if(!file)
//		return;
	SwapFile *file=swapfile;
	
	tile->AllocateBuffer();
	
	int readerror=0;

	readerror=file->ReadAt(tile->GetSwapOffset(),tile->GetData(), tile->BitsLength());
	if(readerror<=0)
		printf("Couldn't read file in SwapTileIn\n");
	
	Unlock();
}

void TileVMM::ResizeFile(SwapFile *file, off_t newsize)
{
//	if(file->GetSize()>newsize)
//		ftruncate(file->GetFD(),newsize);
	file->SetFileSize(newsize);
}

uint64 TileVMM::ObtainOffset(SwapFile *file, uint16 size)
{
/*	Gap *tempgap;
	BList &list=file->GetGaplist();
	uint32 offset;
	
	for(int32 i=0; i<list.CountItems(); i++)
	{
		tempgap=(Gap*)list.ItemAt(i);
		if(!tempgap)
			continue;

		if(tempgap->GetSize()>=size)
		{
			offset=tempgap->start;
			tempgap->start+=size;
			
			if(tempgap->GetSize()==0)
			{
				list.RemoveItem(tempgap);
				delete tempgap;
			}
			return offset;
		}
	}
*/
	uint64 offset;
	GapList &list=file->GetGaplist();

	if(list.Get(size, offset)==B_OK)
		return offset;
	
	// We got this far, so there is no room in the swapfile for the tile and we need to 
	// *make* room

	offset=file->GetFileSize();
	ResizeFile(file,file->GetFileSize()+SWAPFILE_BLOCK_SIZE);
	
/*	if( (offset+size) < file->GetFileSize() )
	{
		Gap *gap=new Gap(offset+size, file->GetFileSize());
		list.AddItem(gap);
	}
*/
	list.Expand(SWAPFILE_BLOCK_SIZE);
	list.Get(size, offset);

	return offset;
}

uint64 TileVMM::CalculateNeededSpace(uint16 w, uint16 h, uint8 bpp)
{
	uint64 rows=(h+MAX_TILE_HEIGHT-1)/ MAX_TILE_HEIGHT;
	uint64 cols=(w+MAX_TILE_WIDTH-1)/ MAX_TILE_WIDTH;
	
	return uint64(rows * cols * MAX_TILE_SIZE * (uint64)bpp);
}
